"""
A reachable game is defined as (G, F) where G = (S_0, S_1, A) induced a
directed graph (S_0 U S_1, A) and F is a subset of S_1. A is a list of pairs
of elements of S_0 U S_1.
"""

import logging


def get_attractor(S0, S1, A, F):
    """
    Compute the attractor set.
    Credit: Dietmar Berwanger in "Graph games with perfect information"
    This algorithm has been modified to not use recursion.
    :param S0: A list of vertices
    :param S1: A list of vertices (must be disjointed of S0)
    :param A: A sub-list (subset) of S0 x S1 U S1 x S0
    :param F: A sub-list (subset) of S1 as list.
    """
    logger = logging.getLogger('main.reachability_game')
    logger.info('"reachability_game.get_attractor" called.')
    
    in_attractor = dict()
    previous = dict()
    num_out_degree = dict()
    S0_set = set(S0)

    for v in S0 + S1:
        in_attractor[v] = False
        previous[v] = set()
        num_out_degree[v] = 0
    
    for u, v in A:
        previous[v].add(u)
        num_out_degree[u] += 1
    
    propagate_stack = list(F)
    while len(propagate_stack) > 0:
        vertex = propagate_stack.pop()
        in_attractor[vertex] = True

        for prev in previous[vertex]:
            num_out_degree[prev] -= 1
            if (prev in S0_set or num_out_degree[prev] == 0) \
                    and not in_attractor[prev]:
                propagate_stack.append(prev)
    
    return [vertex for vertex, is_in_attractor in in_attractor.items()
                if is_in_attractor]


def get_next_winning_moves(current_vertex, A, attractor, player0_move=True):
    """
    Compute a list of next moves that lead to a winning game for the player 0
    if "player0_move" with respect to the game on the graph with the set of
    arcs "A", its attractor set "attractor" and the vertex "current_vertex"
    the token is on.
    :param current_vertex: A vertex
    :param A: A list of arcs
    :param attractor: A list representing the attractor set.
    :param player0_move: A flag meaning that it's player 0's turn to play.
    """
    logger = logging.getLogger('main.reachability_game')
    logger.info('"reachability_game.get_next_winning_moves" called.')

    attractor_set = set(attractor)
    if player0_move:
        return [v for u, v in A if u == current_vertex and v in attractor_set]
    else:
        return [v for u, v in A if u == current_vertex and
            v not in attractor_set]



import sys, argparse, logging, errno
import json, re, itertools
from . import cops_robbers_game as crg


PROGRAM_NAME = 'GGames'
PROGRAM_DESCRIPTION = \
'''This program takes a cop-number problem and decides if it is a k-cop-
winning graph. The graph can be either static or edge periodic. It proceeds
by reducing the problem to a reachability game.'''
VERSION = '1.0'

ERROR_JSON_MSG = \
'''The JSon is not well formatted.'''
ERROR_OPENING_OUTPUT_FILE_MSG = \
'''The output file cannot be opened. Check the permissions of the directory,
or if the file exists and cannot be overwritten.'''
ERROR_OPENING_GRAPH_FILE_MSG = \
'''The file containing the graph cannot be opened. Check if the file exists
and if the permission of reading is granted.'''


def _extract_graph(graph_str):
    """
    Extract the graph from a string.
    :param graph_str: A graph in JSon format.
    """
    json_object = json.loads(graph_str)
    V = json_object['V']; E = list(map(tuple, json_object['E']))

    # Validate that the list of edges does not contain inexistent vertices.
    V_set = set(json_object['V'])
    for u in itertools.chain(*json_object['E']):
        if u not in V_set:
            raise ValueError('Unexpected value detected in \'E\' variable.')
    
    # Validate that every edge has a corresponding binary string.
    if 'tau' in graph_str:
        if len(E) != len(json_object['tau']):
            raise ValueError('Unexpected number of elements in \'tau\' '\
                    'variable.')
        tau = {E[i]: seq for i, seq in enumerate(json_object['tau'])}
        
        # Validate that all elements in 'tau' are composed of binary strings.
        bin_regex = re.compile('^(0*10*)+$')
        for binary_str in tau.values():
            if bin_regex.fullmatch(binary_str) is None:
                raise ValueError('Unexpected value detected in \'tau\' '\
                        'variable.')
        return V, E, tau
    return V, E


def kcop_win(arg_list):
    parser = argparse.ArgumentParser(
            prog=PROGRAM_NAME,
            description=PROGRAM_DESCRIPTION)
    parser.add_argument('k', type=int, help=
            'The number of cops in the game.')
    parser.add_argument('graph_file_path', help=
            'The path to the file containing the description of the graph.')
    parser.add_argument('--output_path', '-o',
            help='Specify the path to the output file containing the presence'\
            ' mapping of the k-cop-win graph.')
    parser.add_argument('--verbose', '-v', action='store_true', help=
            'Output more information.')
    parser.add_argument('--version', action='version', help=
            'Show the current version of the program.',
            version=f'%(prog)s {VERSION}')
    parsed_args = parser.parse_args(args=arg_list)
    
    logger = logging.getLogger('main')
    if parsed_args.verbose:
        # Activate the logger.
        logger.setLevel(logging.INFO)
        logger_handler = logging.StreamHandler(stream=sys.stdout)
        logger_handler.setLevel(logging.INFO)
        logger.addHandler(logger_handler)
    
    if parsed_args.output_path is not None:
        # If an output path is given, it will be used to output the result.
        # Otherwise, the standard output will be used.
        try:
            output = open(parsed_args.output_path, 'w+')
        except OSError as error:
            sys.stderr.write(f'{ERROR_OPENING_OUTPUT_FILE_MSG}\n'\
                    f'{error.strerror}')
            exit(error.errno)
    else:
        output = sys.stdout
    
    # Get the graph
    logger.info('Loading the graph...')
    try:
        file = open(parsed_args.graph_file_path, 'r')
        graph_str = file.read()
        file.close()
        graph = _extract_graph(graph_str)
    except OSError as error:
        sys.stderr.write(f'{ERROR_OPENING_GRAPH_FILE_MSG}\n'\
                f'{error.strerror}')
        exit(error.errno)
    except json.JSONDecodeError as error:
        sys.stderr.write(f'{ERROR_JSON_MSG}\n{error}')
        exit(errno.EINVAL)
    except ValueError as error:
        sys.stderr.write(str(error))
        exit(errno.EINVAL)
    logger.info('Graph loaded.')

    result = crg.is_kcop_win(graph[0], graph[1],
            tau=graph[2] if len(graph) == 3 else None, k=parsed_args.k)
    output.write(f'{result}\n')
    output.flush()
    exit(0)



"""
A Cops and Robbers game is played on an edge periodic (or static) graph
(V, E, tau) where tau is the presence mapping of the edges.
"""


import logging
import math, copy
import functools, itertools
from . import reachability_game



def get_game_graph(V, E, tau=None, k=1):
    """
    Compute the game graph where the "k"-cops and robbers game takes place on
    the edge periodic graph (V, E, tau). If "tau" is not specified, then the
    graph is considered to be static.
    :param V: The list of vertices
    :param E: The list of edges
    :param tau: The presence function of the edges in E in dict
    :param k: The number of cops in the game
    """
    logger = logging.getLogger('main.cops_robbers_game')
    logger.info('"cops_robbers_game.get_game_graph" called.')

    if tau is None: tau = {e: '1' for e in E}

    # Compute an adjacency matrix to simplify the algorithm.
    vertex_index = {u: index for index, u in enumerate(V)}
    adjacency = [[tau[(u, v)] if (u, v) in tau else
                    tau[(v, u)] if (v, u) in tau else
                    '1' if u == v else
                    '0' for u in V] for v in V]
    
    # Compute the least common multiple.
    pattern_lengths = list(map(len, tau.values()))
    time_horizon = functools.reduce(lambda x,y: abs(x*y) // math.gcd(x,y),
            pattern_lengths)

    # Compute the set of vertices of the game graph.
    V_gg = []; A_gg = []
    for t in range(time_horizon):
        for s in [False, True]:
            for *c, r in itertools.product(V, repeat=k+1):
                u = (*c, r, s, t)
                V_gg.append(u)
                if r in c: continue

                next_s = not s
                if s: # Robber's move
                    for next_r in V:
                        edge_pattern = adjacency[vertex_index[r]] \
                                [vertex_index[next_r]]
                        if edge_pattern[t%len(edge_pattern)] == '0' \
                                or next_r in c:
                            continue
                        A_gg.append((u, (*c, next_r, next_s,
                                (t+1)%time_horizon)))
                else: # Cops' move
                    for next_c in itertools.product(V, repeat=k):
                        valid_flag = True # This can be more effective
                        for i in range(len(c)):
                            edge_pattern = adjacency[vertex_index[c[i]]] \
                                    [vertex_index[next_c[i]]]
                            if c[i] != next_c[i] and \
                                    edge_pattern[t%len(edge_pattern)] == '0':
                                # It is impossible for a cops to move in
                                # one rounds to a non adjacent vertex.
                                valid_flag = False
                                break
                        if valid_flag:
                            A_gg.append((u, (*next_c, r, next_s,t)))

    return V_gg, A_gg


def game_graph_to_reachability_game(V_gg, A_gg):
    """
    Compute and return a reachable game corresponding to the game graph
    G = (V_gg, A_gg).
    :param V_gg: A list of vertices of a game graph
    :param A_gg: A list of edges of a game graph
    """
    logger = logging.getLogger('main.com_robber_game')
    logger.info('"cops_robbers_game.game_graph_to_reachability_game" called.')

    S0 = []; S1 = []
    A = copy.deepcopy(A_gg)
    F = []
    for v in V_gg:
        *c, r, s, t = v
        if r in c:
            F.append(v)
        if s:
            S1.append(v)
        else:
            S0.append(v)
    return S0, S1, A, F


def is_kcop_win(V, E, tau=None, k=1):
    """
    Compute if the time-varying graph ("V", "E", "tau") is "k"-cop win.
    :param V: A set of vertices
    :param E: A set of edges
    :param tau: A map from E to a set of bit sequences
    :param k: The number of cops that play on the time-varying graph
    """
    logger = logging.getLogger('main.com_robber_game')
    logger.info('"cops_robbers_game.is_kcop_win" called.')

    attractor = reachability_game.get_attractor(
            *game_graph_to_reachability_game(
            *get_game_graph(V, E, tau, k)))

    n = len(V)
    starting_classes = dict()
    for *c, r, s, t in attractor:
        c = tuple(c)
        if t == 0 and not s:
            if c not in starting_classes:
                starting_classes[c] = set()
            starting_classes[c].add(r)
    for cls in starting_classes.values():
        if len(cls) == n:
            return True
    return False
